@@ -22,4 +22,22 @@
 #define read_barrier()	asm volatile("bcr 15,0" : : : "memory")
 #define write_barrier()	asm volatile("bcr 15,0" : : : "memory")
 
+static inline unsigned long long get_cpu_clock(void)
+{
+	unsigned long long clk;
+
+	__asm__ __volatile__("stck %0" : "=Q" (clk) : : "cc");
+	return clk;
+}
+
+#define ARCH_HAVE_INIT
+extern int tsc_reliable;
+static inline int arch_init(char *envp[])
+{
+	tsc_reliable = 1;
+	return 0;
+}
+
+#define ARCH_HAVE_CPU_CLOCK
+
 #endif
