@@ -1,7 +1,18 @@
 #ifndef ARCH_X86_H
 #define ARCH_X86_H
 
-#define ARCH	(arch_i386)
+static inline void do_cpuid(unsigned int *eax, unsigned int *ebx,
+			    unsigned int *ecx, unsigned int *edx)
+{
+	asm volatile("xchgl %%ebx, %1\ncpuid\nxchgl %%ebx, %1"
+		: "=a" (*eax), "=r" (*ebx), "=c" (*ecx), "=d" (*edx)
+		: "0" (*eax)
+		: "memory");
+}
+
+#include "arch-x86-common.h"
+
+#define FIO_ARCH	(arch_i386)
 
 #ifndef __NR_ioprio_set
 #define __NR_ioprio_set		289
@@ -18,17 +29,8 @@
 #define __NR_sys_vmsplice	316
 #endif
 
-#ifndef __NR_async_exec
-#define __NR_async_exec		325
-#define __NR_async_wait		326
-#define __NR_umem_add		327
-#define __NR_async_thread	328
-#endif
-
 #define	FIO_HUGE_PAGE		4194304
 
-#define FIO_HAVE_SYSLET
-
 #define nop		__asm__ __volatile__("rep;nop": : :"memory")
 #define read_barrier()	__asm__ __volatile__("": : :"memory")
 #define write_barrier()	__asm__ __volatile__("": : :"memory")
@@ -41,7 +43,7 @@ static inline unsigned long arch_ffz(unsigned long bitmask)
 
 static inline unsigned long long get_cpu_clock(void)
 {
-	unsigned long ret;
+	unsigned long long ret;
 
 	__asm__ __volatile__("rdtsc" : "=A" (ret));
 	return ret;
