@@ -27,6 +27,7 @@ extern unsigned int fio_debug_jobno, *fio_debug_jobp;
 #ifdef FIO_INC_DEBUG
 struct debug_level {
 	const char *name;
+	const char *help;
 	unsigned long shift;
 	unsigned int jobno;
 };
@@ -34,7 +35,7 @@ extern struct debug_level debug_levels[];
 
 extern unsigned long fio_debug;
 
-void __dprint(int type, const char *str, ...);
+void __dprint(int type, const char *str, ...) __attribute__((format (printf, 2, 3)));
 
 #define dprint(type, str, args...)			\
 	do {						\
