@@ -1,8 +1,10 @@
 #ifndef FIO_DISKUTIL_H
 #define FIO_DISKUTIL_H
-
+#include "json.h"
 #define FIO_DU_NAME_SZ		64
 
+extern volatile int disk_util_exit;
+
 /*
  * Disk utils as read in /sys/block/<dev>/stat
  */
@@ -40,7 +42,7 @@ struct disk_util {
 
 	char *name;
 	char *sysfs_root;
-	char path[256];
+	char path[PATH_MAX];
 	int major, minor;
 
 	struct disk_util_stat dus;
@@ -94,21 +96,36 @@ static inline void disk_util_dec(struct disk_util *du)
 
 extern struct flist_head disk_list;
 
+extern void wait_for_disk_thread_exit(void);
+
 /*
  * disk util stuff
  */
 #ifdef FIO_HAVE_DISK_UTIL
 extern void print_disk_util(struct disk_util_stat *, struct disk_util_agg *, int terse);
-extern void show_disk_util(int terse);
-extern void free_disk_util(void);
+extern void show_disk_util(int terse, struct json_object *parent);
 extern void init_disk_util(struct thread_data *);
-extern void update_io_ticks(void);
+extern int update_io_ticks(void);
+extern void setup_disk_util(void);
+extern void disk_util_prune_entries(void);
 #else
-#define print_disk_util(dus, agg, terse)
-#define show_disk_util(terse)
-#define free_disk_util()
+static inline void print_disk_util(struct disk_util_stat *du,
+				   struct disk_util_agg *agg, int terse)
+{
+}
+#define show_disk_util(terse, parent)
+#define disk_util_prune_entries()
 #define init_disk_util(td)
-#define update_io_ticks()
+#define setup_disk_util()
+static inline int update_io_ticks(void)
+{
+	return disk_util_exit;
+}
 #endif
 
+static inline void disk_util_start_exit(void)
+{
+	disk_util_exit = 1;
+}
+
 #endif
