@@ -5,16 +5,22 @@
  * Supports both RDMA memory semantics and channel semantics
  *   for the InfiniBand, RoCE and iWARP protocols.
  *
- * This I/O engine is disabled by default. To enable it, execute:
- *
- * $ export EXTFLAGS="-DFIO_HAVE_RDMA"
- * $ export EXTLIBS="-libverbs -lrdmacm"
- *
- * before running make. You will need the Linux RDMA software as well, either
+ * You will need the Linux RDMA software installed, either
  * from your Linux distributor or directly from openfabrics.org:
  *
  * http://www.openfabrics.org/downloads/OFED/
  *
+ * Exchanging steps of RDMA ioengine control messages:
+ *	1. client side sends test mode (RDMA_WRITE/RDMA_READ/SEND)
+ *	   to server side.
+ *	2. server side parses test mode, and sends back confirmation
+ *	   to client side. In RDMA WRITE/READ test, this confirmation
+ *	   includes memory information, such as rkey, address.
+ *	3. client side initiates test loop.
+ *	4. In RDMA WRITE/READ test, client side sends a completion
+ *	   notification to server side. Server side updates its
+ *	   td->done as true.
+ *
  */
 #include <stdio.h>
 #include <stdlib.h>
@@ -30,18 +36,16 @@
 #include <sys/time.h>
 #include <sys/resource.h>
 
-#include <byteswap.h>
 #include <pthread.h>
 #include <inttypes.h>
 
 #include "../fio.h"
-
-#ifdef FIO_HAVE_RDMA
+#include "../hash.h"
 
 #include <rdma/rdma_cma.h>
 #include <infiniband/arch.h>
 
-#define FIO_RDMA_MAX_IO_DEPTH    128
+#define FIO_RDMA_MAX_IO_DEPTH    512
 
 enum rdma_io_mode {
 	FIO_RDMA_UNKNOWN = 0,
@@ -108,6 +112,8 @@ struct rdmaio_data {
 	int io_u_flight_nr;
 	struct io_u **io_us_completed;
 	int io_u_completed_nr;
+
+	struct frand_state rand_state;
 };
 
 static int client_recv(struct thread_data *td, struct ibv_wc *wc)
@@ -311,6 +317,7 @@ static int fio_rdmaio_setup_qp(struct thread_data *td)
 		rd->pd = ibv_alloc_pd(rd->child_cm_id->verbs);
 	else
 		rd->pd = ibv_alloc_pd(rd->cm_id->verbs);
+
 	if (rd->pd == NULL) {
 		log_err("fio: ibv_alloc_pd fail\n");
 		return 1;
@@ -402,7 +409,7 @@ static int fio_rdmaio_setup_control_msg_buffers(struct thread_data *td)
 	/* setup work request */
 	/* recv wq */
 	rd->recv_sgl.addr = (uint64_t) (unsigned long)&rd->recv_buf;
-	rd->recv_sgl.length = sizeof rd->recv_buf;
+	rd->recv_sgl.length = sizeof(rd->recv_buf);
 	rd->recv_sgl.lkey = rd->recv_mr->lkey;
 	rd->rq_wr.sg_list = &rd->recv_sgl;
 	rd->rq_wr.num_sge = 1;
@@ -410,7 +417,7 @@ static int fio_rdmaio_setup_control_msg_buffers(struct thread_data *td)
 
 	/* send wq */
 	rd->send_sgl.addr = (uint64_t) (unsigned long)&rd->send_buf;
-	rd->send_sgl.length = sizeof rd->send_buf;
+	rd->send_sgl.length = sizeof(rd->send_buf);
 	rd->send_sgl.lkey = rd->send_mr->lkey;
 
 	rd->sq_wr.opcode = IBV_WR_SEND;
@@ -427,13 +434,12 @@ static int get_next_channel_event(struct thread_data *td,
 				  enum rdma_cm_event_type wait_event)
 {
 	struct rdmaio_data *rd = td->io_ops->data;
-
-	int ret;
 	struct rdma_cm_event *event;
+	int ret;
 
 	ret = rdma_get_cm_event(channel, &event);
 	if (ret) {
-		log_err("fio: rdma_get_cm_event");
+		log_err("fio: rdma_get_cm_event: %d\n", ret);
 		return 1;
 	}
 
@@ -507,9 +513,9 @@ static struct io_u *fio_rdmaio_event(struct thread_data *td, int event)
 	int i;
 
 	io_u = rd->io_us_completed[0];
-	for (i = 0; i < rd->io_u_completed_nr - 1; i++) {
+	for (i = 0; i < rd->io_u_completed_nr - 1; i++)
 		rd->io_us_completed[i] = rd->io_us_completed[i + 1];
-	}
+
 	rd->io_u_completed_nr--;
 
 	dprint_io_u(io_u, "fio_rdmaio_event");
@@ -521,14 +527,11 @@ static int fio_rdmaio_getevents(struct thread_data *td, unsigned int min,
 				unsigned int max, struct timespec *t)
 {
 	struct rdmaio_data *rd = td->io_ops->data;
-	int r;
 	enum ibv_wc_opcode comp_opcode;
-	comp_opcode = IBV_WC_RDMA_WRITE;
 	struct ibv_cq *ev_cq;
 	void *ev_ctx;
-	int ret;
-
-	r = 0;
+	int ret, r = 0;
+	comp_opcode = IBV_WC_RDMA_WRITE;
 
 	switch (rd->rdma_protocol) {
 	case FIO_RDMA_MEM_WRITE:
@@ -591,7 +594,8 @@ static int fio_rdmaio_send(struct thread_data *td, struct io_u **io_us,
 	enum ibv_wc_opcode comp_opcode;
 	comp_opcode = IBV_WC_RDMA_WRITE;
 #endif
-	int i, index;
+	int i;
+	long index;
 	struct rdma_io_u_data *r_io_u_d;
 
 	r_io_u_d = NULL;
@@ -602,7 +606,7 @@ static int fio_rdmaio_send(struct thread_data *td, struct io_u **io_us,
 		case FIO_RDMA_MEM_WRITE:
 			/* compose work request */
 			r_io_u_d = io_us[i]->engine_data;
-			index = rand() % rd->rmt_nr;
+			index = __rand(&rd->rand_state) % rd->rmt_nr;
 			r_io_u_d->sq_wr.opcode = IBV_WR_RDMA_WRITE;
 			r_io_u_d->sq_wr.wr.rdma.rkey = rd->rmt_us[index].rkey;
 			r_io_u_d->sq_wr.wr.rdma.remote_addr = \
@@ -612,7 +616,7 @@ static int fio_rdmaio_send(struct thread_data *td, struct io_u **io_us,
 		case FIO_RDMA_MEM_READ:
 			/* compose work request */
 			r_io_u_d = io_us[i]->engine_data;
-			index = rand() % rd->rmt_nr;
+			index = __rand(&rd->rand_state) % rd->rmt_nr;
 			r_io_u_d->sq_wr.opcode = IBV_WR_RDMA_READ;
 			r_io_u_d->sq_wr.wr.rdma.rkey = rd->rmt_us[index].rkey;
 			r_io_u_d->sq_wr.wr.rdma.remote_addr = \
@@ -734,11 +738,11 @@ static int fio_rdmaio_commit(struct thread_data *td)
 	io_us = rd->io_us_queued;
 	do {
 		/* RDMA_WRITE or RDMA_READ */
-		if (rd->is_client) {
+		if (rd->is_client)
 			ret = fio_rdmaio_send(td, io_us, rd->io_u_queued_nr);
-		} else if (!rd->is_client) {
+		else if (!rd->is_client)
 			ret = fio_rdmaio_recv(td, io_us, rd->io_u_queued_nr);
-		} else
+		else
 			ret = 0;	/* must be a SYNC */
 
 		if (ret > 0) {
@@ -760,7 +764,7 @@ static int fio_rdmaio_connect(struct thread_data *td, struct fio_file *f)
 	struct rdma_conn_param conn_param;
 	struct ibv_send_wr *bad_wr;
 
-	memset(&conn_param, 0, sizeof conn_param);
+	memset(&conn_param, 0, sizeof(conn_param));
 	conn_param.responder_resources = 1;
 	conn_param.initiator_depth = 1;
 	conn_param.retry_count = 10;
@@ -790,6 +794,16 @@ static int fio_rdmaio_connect(struct thread_data *td, struct fio_file *f)
 	/* wait for remote MR info from server side */
 	rdma_poll_wait(td, IBV_WC_RECV);
 
+	/* In SEND/RECV test, it's a good practice to setup the iodepth of
+	 * of the RECV side deeper than that of the SEND side to
+	 * avoid RNR (receiver not ready) error. The
+	 * SEND side may send so many unsolicited message before
+	 * RECV side commits sufficient recv buffers into recv queue.
+	 * This may lead to RNR error. Here, SEND side pauses for a while
+	 * during which RECV side commits sufficient recv buffers.
+	 */
+	usleep(500000);
+
 	return 0;
 }
 
@@ -800,7 +814,7 @@ static int fio_rdmaio_accept(struct thread_data *td, struct fio_file *f)
 	struct ibv_send_wr *bad_wr;
 
 	/* rdma_accept() - then wait for accept success */
-	memset(&conn_param, 0, sizeof conn_param);
+	memset(&conn_param, 0, sizeof(conn_param));
 	conn_param.responder_resources = 1;
 	conn_param.initiator_depth = 1;
 
@@ -863,17 +877,20 @@ static int fio_rdmaio_close_file(struct thread_data *td, struct fio_file *f)
 		rdma_disconnect(rd->cm_id);
 	else {
 		rdma_disconnect(rd->child_cm_id);
-/*        rdma_disconnect(rd->cm_id); */
+#if 0
+		rdma_disconnect(rd->cm_id);
+#endif
 	}
 
-/*    if (get_next_channel_event(td, rd->cm_channel, RDMA_CM_EVENT_DISCONNECTED) != 0)
-    {
-        log_err("fio: wait for RDMA_CM_EVENT_DISCONNECTED\n");
-        return 1;
-    }*/
+#if 0
+	if (get_next_channel_event(td, rd->cm_channel, RDMA_CM_EVENT_DISCONNECTED) != 0) {
+		log_err("fio: wait for RDMA_CM_EVENT_DISCONNECTED\n");
+		return 1;
+	}
+#endif
 
-	ibv_destroy_qp(rd->qp);
 	ibv_destroy_cq(rd->cq);
+	ibv_destroy_qp(rd->qp);
 
 	if (rd->is_client == 1)
 		rdma_destroy_id(rd->cm_id);
@@ -893,6 +910,7 @@ static int fio_rdmaio_setup_connect(struct thread_data *td, const char *host,
 {
 	struct rdmaio_data *rd = td->io_ops->data;
 	struct ibv_recv_wr *bad_wr;
+	int err;
 
 	rd->addr.sin_family = AF_INET;
 	rd->addr.sin_port = htons(port);
@@ -910,28 +928,28 @@ static int fio_rdmaio_setup_connect(struct thread_data *td, const char *host,
 	}
 
 	/* resolve route */
-	if (rdma_resolve_addr(rd->cm_id, NULL,
-			      (struct sockaddr *)&rd->addr, 2000) != 0) {
-		log_err("fio: rdma_resolve_addr");
+	err = rdma_resolve_addr(rd->cm_id, NULL, (struct sockaddr *)&rd->addr, 2000);
+	if (err != 0) {
+		log_err("fio: rdma_resolve_addr: %d\n", err);
 		return 1;
 	}
 
-	if (get_next_channel_event
-	    (td, rd->cm_channel, RDMA_CM_EVENT_ADDR_RESOLVED)
-	    != 0) {
-		log_err("fio: get_next_channel_event");
+	err = get_next_channel_event(td, rd->cm_channel, RDMA_CM_EVENT_ADDR_RESOLVED);
+	if (err != 0) {
+		log_err("fio: get_next_channel_event: %d\n", err);
 		return 1;
 	}
 
 	/* resolve route */
-	if (rdma_resolve_route(rd->cm_id, 2000) != 0) {
-		log_err("fio: rdma_resolve_route");
+	err = rdma_resolve_route(rd->cm_id, 2000);
+	if (err != 0) {
+		log_err("fio: rdma_resolve_route: %d\n", err);
 		return 1;
 	}
 
-	if (get_next_channel_event
-	    (td, rd->cm_channel, RDMA_CM_EVENT_ROUTE_RESOLVED) != 0) {
-		log_err("fio: get_next_channel_event");
+	err = get_next_channel_event(td, rd->cm_channel, RDMA_CM_EVENT_ROUTE_RESOLVED);
+	if (err != 0) {
+		log_err("fio: get_next_channel_event: %d\n", err);
 		return 1;
 	}
 
@@ -943,8 +961,9 @@ static int fio_rdmaio_setup_connect(struct thread_data *td, const char *host,
 		return 1;
 
 	/* post recv buf */
-	if (ibv_post_recv(rd->qp, &rd->rq_wr, &bad_wr) != 0) {
-		log_err("fio: ibv_post_recv fail\n");
+	err = ibv_post_recv(rd->qp, &rd->rq_wr, &bad_wr);
+	if (err != 0) {
+		log_err("fio: ibv_post_recv fail: %d\n", err);
 		return 1;
 	}
 
@@ -993,24 +1012,11 @@ static int fio_rdmaio_setup_listen(struct thread_data *td, short port)
 	return 0;
 }
 
-static int fio_rdmaio_init(struct thread_data *td)
+static int check_set_rlimits(struct thread_data *td)
 {
-	struct rdmaio_data *rd = td->io_ops->data;
-	unsigned int port;
-	char host[64], buf[128];
-	char *sep, *portp, *modep;
-	int ret;
+#ifdef CONFIG_RLIMIT_MEMLOCK
 	struct rlimit rl;
 
-	if (td_rw(td)) {
-		log_err("fio: rdma connections must be read OR write\n");
-		return 1;
-	}
-	if (td_random(td)) {
-		log_err("fio: RDMA network IO can't be random\n");
-		return 1;
-	}
-
 	/* check RLIMIT_MEMLOCK */
 	if (getrlimit(RLIMIT_MEMLOCK, &rl) != 0) {
 		log_err("fio: getrlimit fail: %d(%s)\n",
@@ -1035,6 +1041,31 @@ static int fio_rdmaio_init(struct thread_data *td)
 			return 1;
 		}
 	}
+#endif
+
+	return 0;
+}
+
+static int fio_rdmaio_init(struct thread_data *td)
+{
+	struct rdmaio_data *rd = td->io_ops->data;
+	unsigned int max_bs;
+	unsigned int port;
+	char host[64], buf[128];
+	char *sep, *portp, *modep;
+	int ret, i;
+
+	if (td_rw(td)) {
+		log_err("fio: rdma connections must be read OR write\n");
+		return 1;
+	}
+	if (td_random(td)) {
+		log_err("fio: RDMA network IO can't be random\n");
+		return 1;
+	}
+
+	if (check_set_rlimits(td))
+		return 1;
 
 	strcpy(buf, td->o.filename);
 
@@ -1119,13 +1150,10 @@ static int fio_rdmaio_init(struct thread_data *td)
 		ret = fio_rdmaio_setup_connect(td, host, port);
 	}
 
-	struct flist_head *entry;
-	unsigned int max_bs;
 	max_bs = max(td->o.max_bs[DDIR_READ], td->o.max_bs[DDIR_WRITE]);
 	/* register each io_u in the free list */
-	int i = 0;
-	flist_for_each(entry, &td->io_u_freelist) {
-		struct io_u *io_u = flist_entry(entry, struct io_u, list);
+	for (i = 0; i < td->io_u_freelist.nr; i++) {
+		struct io_u *io_u = td->io_u_freelist.io_us[i];
 
 		io_u->engine_data = malloc(sizeof(struct rdma_io_u_data));
 		memset(io_u->engine_data, 0, sizeof(struct rdma_io_u_data));
@@ -1145,9 +1173,9 @@ static int fio_rdmaio_init(struct thread_data *td)
 		rd->send_buf.rmt_us[i].rkey = htonl(io_u->mr->rkey);
 		rd->send_buf.rmt_us[i].size = htonl(max_bs);
 
-/*    log_info("fio: Send rkey %x addr %" PRIx64 " len %d to client\n",
-          io_u->mr->rkey, io_u->buf, max_bs); */
-		i++;
+#if 0
+		log_info("fio: Send rkey %x addr %" PRIx64 " len %d to client\n", io_u->mr->rkey, io_u->buf, max_bs); */
+#endif
 	}
 
 	rd->send_buf.nr = htonl(i);
@@ -1162,16 +1190,8 @@ static void fio_rdmaio_cleanup(struct thread_data *td)
 {
 	struct rdmaio_data *rd = td->io_ops->data;
 
-	if (rd) {
-/*        if (nd->listenfd != -1)
-            close(nd->listenfd);
-        if (nd->pipes[0] != -1)
-            close(nd->pipes[0]);
-        if (nd->pipes[1] != -1)
-            close(nd->pipes[1]);
-*/
+	if (rd)
 		free(rd);
-	}
 }
 
 static int fio_rdmaio_setup(struct thread_data *td)
@@ -1179,9 +1199,10 @@ static int fio_rdmaio_setup(struct thread_data *td)
 	struct rdmaio_data *rd;
 
 	if (!td->io_ops->data) {
-		rd = malloc(sizeof(*rd));;
+		rd = malloc(sizeof(*rd));
 
 		memset(rd, 0, sizeof(*rd));
+		init_rand_seed(&rd->rand_state, (unsigned int) GOLDEN_RATIO_PRIME);
 		td->io_ops->data = rd;
 	}
 
@@ -1204,49 +1225,6 @@ static struct ioengine_ops ioengine_rw = {
 	.flags		= FIO_DISKLESSIO | FIO_UNIDIR | FIO_PIPEIO,
 };
 
-#else /* FIO_HAVE_RDMA */
-
-static int fio_rdmaio_open_file(struct thread_data *td, struct fio_file *f)
-{
-	return 0;
-}
-
-static int fio_rdmaio_close_file(struct thread_data *td, struct fio_file *f)
-{
-	return 0;
-}
-
-static int fio_rdmaio_queue(struct thread_data *td, struct io_u *io_u)
-{
-	return FIO_Q_COMPLETED;
-}
-
-static int fio_rdmaio_init(struct thread_data fio_unused * td)
-{
-	log_err("fio: rdma(librdmacm libibverbs) not available\n");
-	log_err("     You haven't compiled rdma ioengine into fio.\n");
-	log_err("     If you want to try rdma ioengine,\n");
-	log_err("     make sure OFED is installed,\n");
-	log_err("     $ ofed_info\n");
-	log_err("     then try to make fio as follows:\n");
-	log_err("     $ export EXTFLAGS=\"-DFIO_HAVE_RDMA\"\n");
-	log_err("     $ export EXTLIBS=\"-libverbs -lrdmacm\"\n");
-	log_err("     $ make clean && make\n");
-	return 1;
-}
-
-static struct ioengine_ops ioengine_rw = {
-	.name		= "rdma",
-	.version	= FIO_IOOPS_VERSION,
-	.init		= fio_rdmaio_init,
-	.queue		= fio_rdmaio_queue,
-	.open_file	= fio_rdmaio_open_file,
-	.close_file	= fio_rdmaio_close_file,
-	.flags		= FIO_SYNCIO | FIO_DISKLESSIO | FIO_UNIDIR | FIO_PIPEIO,
-};
-
-#endif
-
 static void fio_init fio_rdmaio_register(void)
 {
 	register_ioengine(&ioengine_rw);
