@@ -14,6 +14,11 @@
 
 #include "../fio.h"
 
+/*
+ * Sync engine uses engine_data to store last offset
+ */
+#define LAST_POS(f)	((f)->engine_data)
+
 struct syncio_data {
 	struct iovec *iovecs;
 	struct io_u **io_us;
@@ -33,7 +38,7 @@ static int fio_syncio_prep(struct thread_data *td, struct io_u *io_u)
 	if (!ddir_rw(io_u->ddir))
 		return 0;
 
-	if (f->file_pos != -1ULL && f->file_pos == io_u->offset)
+	if (LAST_POS(f) != -1ULL && LAST_POS(f) == io_u->offset)
 		return 0;
 
 	if (lseek(f->fd, io_u->offset, SEEK_SET) == -1) {
@@ -47,7 +52,7 @@ static int fio_syncio_prep(struct thread_data *td, struct io_u *io_u)
 static int fio_io_end(struct thread_data *td, struct io_u *io_u, int ret)
 {
 	if (io_u->file && ret >= 0 && ddir_rw(io_u->ddir))
-		io_u->file->file_pos = io_u->offset + ret;
+		LAST_POS(io_u->file) = io_u->offset + ret;
 
 	if (ret != (int) io_u->xfer_buflen) {
 		if (ret >= 0) {
@@ -64,6 +69,33 @@ static int fio_io_end(struct thread_data *td, struct io_u *io_u, int ret)
 	return FIO_Q_COMPLETED;
 }
 
+#ifdef CONFIG_PWRITEV
+static int fio_pvsyncio_queue(struct thread_data *td, struct io_u *io_u)
+{
+	struct syncio_data *sd = td->io_ops->data;
+	struct iovec *iov = &sd->iovecs[0];
+	struct fio_file *f = io_u->file;
+	int ret;
+
+	fio_ro_check(td, io_u);
+
+	iov->iov_base = io_u->xfer_buf;
+	iov->iov_len = io_u->xfer_buflen;
+
+	if (io_u->ddir == DDIR_READ)
+		ret = preadv(f->fd, iov, 1, io_u->offset);
+	else if (io_u->ddir == DDIR_WRITE)
+		ret = pwritev(f->fd, iov, 1, io_u->offset);
+	else if (io_u->ddir == DDIR_TRIM) {
+		do_io_u_trim(td, io_u);
+		return FIO_Q_COMPLETED;
+	} else
+		ret = do_io_u_sync(td, io_u);
+
+	return fio_io_end(td, io_u, ret);
+}
+#endif
+
 static int fio_psyncio_queue(struct thread_data *td, struct io_u *io_u)
 {
 	struct fio_file *f = io_u->file;
@@ -75,9 +107,10 @@ static int fio_psyncio_queue(struct thread_data *td, struct io_u *io_u)
 		ret = pread(f->fd, io_u->xfer_buf, io_u->xfer_buflen, io_u->offset);
 	else if (io_u->ddir == DDIR_WRITE)
 		ret = pwrite(f->fd, io_u->xfer_buf, io_u->xfer_buflen, io_u->offset);
-	else if (io_u->ddir == DDIR_TRIM)
-		ret = do_io_u_trim(td, io_u);
-	else
+	else if (io_u->ddir == DDIR_TRIM) {
+		do_io_u_trim(td, io_u);
+		return FIO_Q_COMPLETED;
+	} else
 		ret = do_io_u_sync(td, io_u);
 
 	return fio_io_end(td, io_u, ret);
@@ -94,9 +127,10 @@ static int fio_syncio_queue(struct thread_data *td, struct io_u *io_u)
 		ret = read(f->fd, io_u->xfer_buf, io_u->xfer_buflen);
 	else if (io_u->ddir == DDIR_WRITE)
 		ret = write(f->fd, io_u->xfer_buf, io_u->xfer_buflen);
-	else if (io_u->ddir == DDIR_TRIM)
-		ret = do_io_u_trim(td, io_u);
-	else
+	else if (io_u->ddir == DDIR_TRIM) {
+		do_io_u_trim(td, io_u);
+		return FIO_Q_COMPLETED;
+	} else
 		ret = do_io_u_sync(td, io_u);
 
 	return fio_io_end(td, io_u, ret);
@@ -322,11 +356,28 @@ static struct ioengine_ops ioengine_vrw = {
 	.flags		= FIO_SYNCIO,
 };
 
+#ifdef CONFIG_PWRITEV
+static struct ioengine_ops ioengine_pvrw = {
+	.name		= "pvsync",
+	.version	= FIO_IOOPS_VERSION,
+	.init		= fio_vsyncio_init,
+	.cleanup	= fio_vsyncio_cleanup,
+	.queue		= fio_pvsyncio_queue,
+	.open_file	= generic_open_file,
+	.close_file	= generic_close_file,
+	.get_file_size	= generic_get_file_size,
+	.flags		= FIO_SYNCIO,
+};
+#endif
+
 static void fio_init fio_syncio_register(void)
 {
 	register_ioengine(&ioengine_rw);
 	register_ioengine(&ioengine_prw);
 	register_ioengine(&ioengine_vrw);
+#ifdef CONFIG_PWRITEV
+	register_ioengine(&ioengine_pvrw);
+#endif
 }
 
 static void fio_exit fio_syncio_unregister(void)
@@ -334,4 +385,7 @@ static void fio_exit fio_syncio_unregister(void)
 	unregister_ioengine(&ioengine_rw);
 	unregister_ioengine(&ioengine_prw);
 	unregister_ioengine(&ioengine_vrw);
+#ifdef CONFIG_PWRITEV
+	unregister_ioengine(&ioengine_pvrw);
+#endif
 }
