@@ -14,12 +14,10 @@
 #include <inttypes.h>
 #include <assert.h>
 
-struct thread_data;
-
 #include "compiler/compiler.h"
+#include "thread_options.h"
 #include "flist.h"
 #include "fifo.h"
-#include "rbtree.h"
 #include "arch/arch.h"
 #include "os/os.h"
 #include "mutex.h"
@@ -36,28 +34,26 @@ struct thread_data;
 #include "gettime.h"
 #include "lib/getopt.h"
 #include "lib/rand.h"
+#include "lib/rbtree.h"
+#include "client.h"
 #include "server.h"
 #include "stat.h"
 #include "flow.h"
+#include "io_u_queue.h"
 
-#ifdef FIO_HAVE_GUASI
-#include <guasi.h>
-#endif
-
-#ifdef FIO_HAVE_SOLARISAIO
+#ifdef CONFIG_SOLARISAIO
 #include <sys/asynch.h>
 #endif
 
+#ifdef CONFIG_LIBNUMA
+#include <linux/mempolicy.h>
+#include <numa.h>
+
 /*
- * What type of allocation to use for io buffers
+ * "local" is pseudo-policy
  */
-enum fio_memtype {
-	MEM_MALLOC = 0,	/* ordinary malloc */
-	MEM_SHM,	/* use shared memory segments */
-	MEM_SHMHUGE,	/* use shared memory segments with huge pages */
-	MEM_MMAP,	/* use anonynomous mmap */
-	MEM_MMAPHUGE,	/* memory mapped huge file */
-};
+#define MPOL_LOCAL MPOL_MAX
+#endif
 
 /*
  * offset generator types
@@ -67,209 +63,29 @@ enum {
 	RW_SEQ_IDENT,
 };
 
-/*
- * What type of errors to continue on when continue_on_error is used
- */
-enum error_type {
-        ERROR_TYPE_NONE = 0,
-        ERROR_TYPE_READ = 1 << 0,
-        ERROR_TYPE_WRITE = 1 << 1,
-        ERROR_TYPE_VERIFY = 1 << 2,
-        ERROR_TYPE_ANY = 0xffff,
+enum {
+	TD_F_VER_BACKLOG	= 1,
+	TD_F_TRIM_BACKLOG	= 2,
+	TD_F_READ_IOLOG		= 4,
+	TD_F_REFILL_BUFFERS	= 8,
+	TD_F_SCRAMBLE_BUFFERS	= 16,
+	TD_F_VER_NONE		= 32,
+	TD_F_PROFILE_OPS	= 64,
 };
 
-struct bssplit {
-	unsigned int bs;
-	unsigned char perc;
-};
-
-struct thread_options {
-	int pad;
-	char *description;
-	char *name;
-	char *directory;
-	char *filename;
-	char *opendir;
-	char *ioengine;
-	enum td_ddir td_ddir;
-	unsigned int rw_seq;
-	unsigned int kb_base;
-	unsigned int ddir_seq_nr;
-	long ddir_seq_add;
-	unsigned int iodepth;
-	unsigned int iodepth_low;
-	unsigned int iodepth_batch;
-	unsigned int iodepth_batch_complete;
-
-	unsigned long long size;
-	unsigned int size_percent;
-	unsigned int fill_device;
-	unsigned long long file_size_low;
-	unsigned long long file_size_high;
-	unsigned long long start_offset;
-
-	unsigned int bs[2];
-	unsigned int ba[2];
-	unsigned int min_bs[2];
-	unsigned int max_bs[2];
-	struct bssplit *bssplit[2];
-	unsigned int bssplit_nr[2];
-
-	unsigned int nr_files;
-	unsigned int open_files;
-	enum file_lock_mode file_lock_mode;
-	unsigned int lockfile_batch;
-
-	unsigned int odirect;
-	unsigned int invalidate_cache;
-	unsigned int create_serialize;
-	unsigned int create_fsync;
-	unsigned int create_on_open;
-	unsigned int create_only;
-	unsigned int end_fsync;
-	unsigned int pre_read;
-	unsigned int sync_io;
-	unsigned int verify;
-	unsigned int do_verify;
-	unsigned int verifysort;
-	unsigned int verify_interval;
-	unsigned int verify_offset;
-	char verify_pattern[MAX_PATTERN_SIZE];
-	unsigned int verify_pattern_bytes;
-	unsigned int verify_fatal;
-	unsigned int verify_dump;
-	unsigned int verify_async;
-	unsigned long long verify_backlog;
-	unsigned int verify_batch;
-	unsigned int use_thread;
-	unsigned int unlink;
-	unsigned int do_disk_util;
-	unsigned int override_sync;
-	unsigned int rand_repeatable;
-	unsigned int use_os_rand;
-	unsigned int write_lat_log;
-	unsigned int write_bw_log;
-	unsigned int write_iops_log;
-	unsigned int log_avg_msec;
-	unsigned int norandommap;
-	unsigned int softrandommap;
-	unsigned int bs_unaligned;
-	unsigned int fsync_on_close;
-
-	unsigned int hugepage_size;
-	unsigned int rw_min_bs;
-	unsigned int thinktime;
-	unsigned int thinktime_spin;
-	unsigned int thinktime_blocks;
-	unsigned int fsync_blocks;
-	unsigned int fdatasync_blocks;
-	unsigned int barrier_blocks;
-	unsigned long long start_delay;
-	unsigned long long timeout;
-	unsigned long long ramp_time;
-	unsigned int overwrite;
-	unsigned int bw_avg_time;
-	unsigned int iops_avg_time;
-	unsigned int loops;
-	unsigned long long zone_range;
-	unsigned long long zone_size;
-	unsigned long long zone_skip;
-	enum fio_memtype mem_type;
-	unsigned int mem_align;
-
-	unsigned int stonewall;
-	unsigned int new_group;
-	unsigned int numjobs;
-	os_cpu_mask_t cpumask;
-	unsigned int cpumask_set;
-	os_cpu_mask_t verify_cpumask;
-	unsigned int verify_cpumask_set;
-	unsigned int iolog;
-	unsigned int rwmixcycle;
-	unsigned int rwmix[2];
-	unsigned int nice;
-	unsigned int file_service_type;
-	unsigned int group_reporting;
-	unsigned int fadvise_hint;
-	enum fio_fallocate_mode fallocate_mode;
-	unsigned int zero_buffers;
-	unsigned int refill_buffers;
-	unsigned int scramble_buffers;
-	unsigned int compress_percentage;
-	unsigned int compress_chunk;
-	unsigned int time_based;
-	unsigned int disable_lat;
-	unsigned int disable_clat;
-	unsigned int disable_slat;
-	unsigned int disable_bw;
-	unsigned int gtod_reduce;
-	unsigned int gtod_cpu;
-	unsigned int gtod_offload;
-	enum fio_cs clocksource;
-	unsigned int no_stall;
-	unsigned int trim_percentage;
-	unsigned int trim_batch;
-	unsigned int trim_zero;
-	unsigned long long trim_backlog;
-	unsigned int clat_percentiles;
-	unsigned int overwrite_plist;
-	fio_fp64_t percentile_list[FIO_IO_U_LIST_MAX_LEN];
-
-	char *read_iolog_file;
-	char *write_iolog_file;
-	char *bw_log_file;
-	char *lat_log_file;
-	char *iops_log_file;
-	char *replay_redirect;
-
-	/*
-	 * Pre-run and post-run shell
-	 */
-	char *exec_prerun;
-	char *exec_postrun;
-
-	unsigned int rate[2];
-	unsigned int ratemin[2];
-	unsigned int ratecycle;
-	unsigned int rate_iops[2];
-	unsigned int rate_iops_min[2];
-
-	char *ioscheduler;
-
-	/*
-	 * CPU "io" cycle burner
-	 */
-	unsigned int cpuload;
-	unsigned int cpucycle;
-
-	/*
-	 * I/O Error handling
-	 */
-	enum error_type continue_on_error;
-
-	/*
-	 * Benchmark profile type
-	 */
-	char *profile;
-
-	/*
-	 * blkio cgroup support
-	 */
-	char *cgroup;
-	unsigned int cgroup_weight;
-	unsigned int cgroup_nodelete;
-
-	unsigned int uid;
-	unsigned int gid;
-
-	int flow_id;
-	int flow;
-	int flow_watermark;
-	unsigned int flow_sleep;
-
-	unsigned long long offset_increment;
-
-	unsigned int sync_file_range;
+enum {
+	FIO_RAND_BS_OFF		= 0,
+	FIO_RAND_VER_OFF,
+	FIO_RAND_MIX_OFF,
+	FIO_RAND_FILE_OFF,
+	FIO_RAND_BLOCK_OFF,
+	FIO_RAND_FILE_SIZE_OFF,
+	FIO_RAND_TRIM_OFF,
+	FIO_RAND_BUF_OFF,
+	FIO_RAND_SEQ_RAND_READ_OFF,
+	FIO_RAND_SEQ_RAND_WRITE_OFF,
+	FIO_RAND_SEQ_RAND_TRIM_OFF,
+	FIO_RAND_NR_OFFS,
 };
 
 /*
@@ -277,29 +93,35 @@ struct thread_options {
  */
 struct thread_data {
 	struct thread_options o;
+	unsigned long flags;
 	void *eo;
 	char verror[FIO_VERROR_SIZE];
 	pthread_t thread;
-	int thread_number;
-	int groupid;
+	unsigned int thread_number;
+	unsigned int groupid;
 	struct thread_stat ts;
 
+	int client_type;
+
 	struct io_log *slat_log;
 	struct io_log *clat_log;
 	struct io_log *lat_log;
 	struct io_log *bw_log;
 	struct io_log *iops_log;
 
-	uint64_t stat_io_bytes[2];
+	uint64_t stat_io_bytes[DDIR_RWDIR_CNT];
 	struct timeval bw_sample_time;
 
-	uint64_t stat_io_blocks[2];
+	uint64_t stat_io_blocks[DDIR_RWDIR_CNT];
 	struct timeval iops_sample_time;
 
+	volatile int update_rusage;
+	struct fio_mutex *rusage_sem;
 	struct rusage ru_start;
 	struct rusage ru_end;
 
 	struct fio_file **files;
+	unsigned char *file_locks;
 	unsigned int files_size;
 	unsigned int files_index;
 	unsigned int nr_open_files;
@@ -318,12 +140,9 @@ struct thread_data {
 	size_t orig_buffer_size;
 	volatile int terminate;
 	volatile int runstate;
-	unsigned int ioprio;
-	unsigned int ioprio_set;
 	unsigned int last_was_sync;
 	enum fio_ddir last_ddir;
 
-	char *mmapfile;
 	int mmapfd;
 
 	void *iolog_buf;
@@ -331,7 +150,7 @@ struct thread_data {
 
 	char *sysfs_root;
 
-	unsigned long rand_seeds[8];
+	unsigned long rand_seeds[FIO_RAND_NR_OFFS];
 
 	union {
 		os_random_state_t bsrange_state;
@@ -377,9 +196,9 @@ struct thread_data {
 	/*
 	 * List of free and busy io_u's
 	 */
-	struct flist_head io_u_freelist;
-	struct flist_head io_u_busylist;
-	struct flist_head io_u_requeues;
+	struct io_u_ring io_u_requeues;
+	struct io_u_queue io_u_freelist;
+	struct io_u_queue io_u_all;
 	pthread_mutex_t io_u_lock;
 	pthread_cond_t free_cond;
 
@@ -395,22 +214,22 @@ struct thread_data {
 	/*
 	 * Rate state
 	 */
-	unsigned long long rate_bps[2];
-	long rate_pending_usleep[2];
-	unsigned long rate_bytes[2];
-	unsigned long rate_blocks[2];
-	struct timeval lastrate[2];
+	uint64_t rate_bps[DDIR_RWDIR_CNT];
+	long rate_pending_usleep[DDIR_RWDIR_CNT];
+	unsigned long rate_bytes[DDIR_RWDIR_CNT];
+	unsigned long rate_blocks[DDIR_RWDIR_CNT];
+	struct timeval lastrate[DDIR_RWDIR_CNT];
 
-	unsigned long long total_io_size;
-	unsigned long long fill_device_size;
+	uint64_t total_io_size;
+	uint64_t fill_device_size;
 
-	unsigned long io_issues[2];
-	unsigned long long io_blocks[2];
-	unsigned long long this_io_blocks[2];
-	unsigned long long io_bytes[2];
-	unsigned long long io_skip_bytes;
-	unsigned long long this_io_bytes[2];
-	unsigned long long zone_bytes;
+	unsigned long io_issues[DDIR_RWDIR_CNT];
+	uint64_t io_blocks[DDIR_RWDIR_CNT];
+	uint64_t this_io_blocks[DDIR_RWDIR_CNT];
+	uint64_t io_bytes[DDIR_RWDIR_CNT];
+	uint64_t io_skip_bytes;
+	uint64_t this_io_bytes[DDIR_RWDIR_CNT];
+	uint64_t zone_bytes;
 	struct fio_mutex *mutex;
 
 	/*
@@ -440,6 +259,14 @@ struct thread_data {
 	enum fio_ddir rwmix_ddir;
 	unsigned int ddir_seq_nr;
 
+	/*
+	 * rand/seq mixed workload state
+	 */
+	union {
+		os_random_state_t seq_rand_state[DDIR_RWDIR_CNT];
+		struct frand_state __seq_rand_state[DDIR_RWDIR_CNT];
+	};
+
 	/*
 	 * IO history logs for verification. We use a tree for sorting,
 	 * if we are overwriting. Otherwise just use a fifo.
@@ -459,6 +286,8 @@ struct thread_data {
 	struct flist_head trim_list;
 	unsigned long trim_entries;
 
+	struct flist_head next_rand_list;
+
 	/*
 	 * for fileservice, how often to switch to a new file
 	 */
@@ -489,6 +318,8 @@ struct thread_data {
 	 */
 	struct prof_io_ops prof_io_ops;
 	void *prof_data;
+
+	void *pinned_mem;
 };
 
 /*
@@ -502,12 +333,12 @@ enum {
 
 #define __td_verror(td, err, msg, func)					\
 	do {								\
+		int e = (err);						\
 		if ((td)->error)					\
 			break;						\
-		int e = (err);						\
 		(td)->error = e;					\
 		if (!(td)->first_error)					\
-			snprintf(td->verror, sizeof(td->verror) - 1, "file:%s:%d, func=%s, error=%s", __FILE__, __LINE__, (func), (msg));		\
+			snprintf(td->verror, sizeof(td->verror), "file:%s:%d, func=%s, error=%s", __FILE__, __LINE__, (func), (msg));		\
 	} while (0)
 
 
@@ -523,27 +354,28 @@ enum {
 
 extern int exitall_on_terminate;
 extern unsigned int thread_number;
-extern unsigned int nr_process, nr_thread;
+extern unsigned int stat_number;
 extern int shm_id;
 extern int groupid;
-extern int terse_output;
+extern int output_format;
 extern int temp_stall_ts;
-extern unsigned long long mlock_size;
 extern uintptr_t page_mask, page_size;
 extern int read_only;
 extern int eta_print;
+extern int eta_new_line;
 extern unsigned long done_secs;
 extern char *job_section;
 extern int fio_gtod_offload;
 extern int fio_gtod_cpu;
 extern enum fio_cs fio_clock_source;
+extern int fio_clock_source_set;
 extern int warnings_fatal;
 extern int terse_version;
 extern int is_backend;
 extern int nr_clients;
 extern int log_syslog;
+extern int status_interval;
 extern const char fio_version_string[];
-extern const fio_fp64_t def_percentile_list[FIO_IO_U_LIST_MAX_LEN];
 
 extern struct thread_data *threads;
 
@@ -552,37 +384,12 @@ static inline void fio_ro_check(struct thread_data *td, struct io_u *io_u)
 	assert(!(io_u->ddir == DDIR_WRITE && !td_write(td)));
 }
 
-#define BLOCKS_PER_MAP		(8 * sizeof(unsigned long))
-#define TO_MAP_BLOCK(f, b)	(b)
-#define RAND_MAP_IDX(f, b)	(TO_MAP_BLOCK(f, b) / BLOCKS_PER_MAP)
-#define RAND_MAP_BIT(f, b)	(TO_MAP_BLOCK(f, b) & (BLOCKS_PER_MAP - 1))
-
 #define REAL_MAX_JOBS		2048
 
-#define td_non_fatal_error(e)	((e) == EIO || (e) == EILSEQ)
-
-static inline enum error_type td_error_type(enum fio_ddir ddir, int err)
-{
-	if (err == EILSEQ)
-		return ERROR_TYPE_VERIFY;
-	if (ddir == DDIR_READ)
-		return ERROR_TYPE_READ;
-	return ERROR_TYPE_WRITE;
-}
-
-static inline void update_error_count(struct thread_data *td, int err)
-{
-	td->total_err_count++;
-	if (td->total_err_count == 1)
-		td->first_error = err;
-}
-
 static inline int should_fsync(struct thread_data *td)
 {
 	if (td->last_was_sync)
 		return 0;
-	if (td->o.odirect)
-		return 0;
 	if (td_write(td) || td_rw(td) || td->o.override_sync)
 		return 1;
 
@@ -592,9 +399,10 @@ static inline int should_fsync(struct thread_data *td)
 /*
  * Init/option functions
  */
+extern int __must_check fio_init_options(void);
 extern int __must_check parse_options(int, char **);
-extern int parse_jobs_ini(char *, int, int);
-extern int parse_cmd_line(int, char **);
+extern int parse_jobs_ini(char *, int, int, int);
+extern int parse_cmd_line(int, char **, int);
 extern int fio_backend(void);
 extern void reset_fio_state(void);
 extern void clear_io_state(struct thread_data *);
@@ -609,9 +417,14 @@ extern void fio_options_dup_and_init(struct option *);
 extern void fio_options_mem_dupe(struct thread_data *);
 extern void options_mem_dupe(void *data, struct fio_option *options);
 extern void td_fill_rand_seeds(struct thread_data *);
-extern void add_job_opts(const char **);
-extern char *num2str(unsigned long, int, int, int);
+extern void add_job_opts(const char **, int);
+extern char *num2str(unsigned long, int, int, int, int);
 extern int ioengine_load(struct thread_data *);
+extern int parse_dryrun(void);
+
+extern uintptr_t page_mask;
+extern uintptr_t page_size;
+extern int initialize_fio(char *envp[]);
 
 #define FIO_GETOPT_JOB		0x89000000
 #define FIO_GETOPT_IOENGINE	0x98000000
@@ -622,6 +435,7 @@ extern int ioengine_load(struct thread_data *);
  */
 extern void print_thread_status(void);
 extern void print_status_init(int);
+extern char *fio_uint_to_kmg(unsigned int val);
 
 /*
  * Thread life cycle. Once a thread has a runstate beyond TD_INITIALIZED, it
@@ -634,6 +448,7 @@ enum {
 	TD_CREATED,
 	TD_INITIALIZED,
 	TD_RAMP,
+	TD_SETTING_UP,
 	TD_RUNNING,
 	TD_PRE_READING,
 	TD_VERIFYING,
@@ -649,10 +464,11 @@ extern void fio_terminate_threads(int);
 /*
  * Memory helpers
  */
-extern int __must_check fio_pin_memory(void);
-extern void fio_unpin_memory(void);
+extern int __must_check fio_pin_memory(struct thread_data *);
+extern void fio_unpin_memory(struct thread_data *);
 extern int __must_check allocate_io_mem(struct thread_data *);
 extern void free_io_mem(struct thread_data *);
+extern void free_threads_shm(void);
 
 /*
  * Reset stats after ramp time completes
@@ -667,13 +483,6 @@ extern int is_blktrace(const char *);
 extern int load_blktrace(struct thread_data *, const char *);
 #endif
 
-/*
- * Mark unused variables passed to ops functions as unused, to silence gcc
- */
-#define fio_unused	__attribute((__unused__))
-#define fio_init	__attribute__((constructor))
-#define fio_exit	__attribute__((destructor))
-
 #define for_each_td(td, i)	\
 	for ((i) = 0, (td) = &threads[0]; (i) < (int) thread_number; (i)++, (td)++)
 #define for_each_file(td, f, i)	\
@@ -717,18 +526,36 @@ static inline int __should_check_rate(struct thread_data *td,
 }
 
 static inline int should_check_rate(struct thread_data *td,
-				    unsigned long *bytes_done)
+				    uint64_t *bytes_done)
 {
 	int ret = 0;
 
-	if (bytes_done[0])
-		ret |= __should_check_rate(td, 0);
-	if (bytes_done[1])
-		ret |= __should_check_rate(td, 1);
+	if (bytes_done[DDIR_READ])
+		ret |= __should_check_rate(td, DDIR_READ);
+	if (bytes_done[DDIR_WRITE])
+		ret |= __should_check_rate(td, DDIR_WRITE);
+	if (bytes_done[DDIR_TRIM])
+		ret |= __should_check_rate(td, DDIR_TRIM);
 
 	return ret;
 }
 
+static inline unsigned int td_max_bs(struct thread_data *td)
+{
+	unsigned int max_bs;
+
+	max_bs = max(td->o.max_bs[DDIR_READ], td->o.max_bs[DDIR_WRITE]);
+	return max(td->o.max_bs[DDIR_TRIM], max_bs);
+}
+
+static inline unsigned int td_min_bs(struct thread_data *td)
+{
+	unsigned int min_bs;
+
+	min_bs = min(td->o.min_bs[DDIR_READ], td->o.min_bs[DDIR_WRITE]);
+	return min(td->o.min_bs[DDIR_TRIM], min_bs);
+}
+
 static inline int is_power_of_2(unsigned int val)
 {
 	return (val != 0 && ((val & (val - 1)) == 0));
@@ -759,4 +586,23 @@ static inline void td_io_u_free_notify(struct thread_data *td)
 extern const char *fio_get_arch_string(int);
 extern const char *fio_get_os_string(int);
 
+#define ARRAY_SIZE(x) (sizeof((x)) / (sizeof((x)[0])))
+
+enum {
+	FIO_OUTPUT_TERSE	= 0,
+	FIO_OUTPUT_JSON,
+	FIO_OUTPUT_NORMAL,
+};
+
+enum {
+	FIO_RAND_DIST_RANDOM	= 0,
+	FIO_RAND_DIST_ZIPF,
+	FIO_RAND_DIST_PARETO,
+};
+
+enum {
+	FIO_RAND_GEN_TAUSWORTHE = 0,
+	FIO_RAND_GEN_LFSR,
+};
+
 #endif
