@@ -28,7 +28,9 @@
 #error Define GOLDEN_RATIO_PRIME for your wordsize.
 #endif
 
-static inline unsigned long hash_long(unsigned long val, unsigned int bits)
+#define GR_PRIME_64	0x9e37fffffffc0001ULL
+
+static inline unsigned long __hash_long(unsigned long val)
 {
 	unsigned long hash = val;
 
@@ -52,8 +54,18 @@ static inline unsigned long hash_long(unsigned long val, unsigned int bits)
 	hash *= GOLDEN_RATIO_PRIME;
 #endif
 
+	return hash;
+}
+
+static inline unsigned long hash_long(unsigned long val, unsigned int bits)
+{
 	/* High bits are more random, so use them. */
-	return hash >> (BITS_PER_LONG - bits);
+	return __hash_long(val) >> (BITS_PER_LONG - bits);
+}
+
+static inline uint64_t __hash_u64(uint64_t val)
+{
+	return val * GR_PRIME_64;
 }
 	
 static inline unsigned long hash_ptr(void *ptr, unsigned int bits)
