@@ -6,6 +6,8 @@
 #include "../flist.h"
 
 FILE *f_err;
+struct timeval *fio_tv = NULL;
+unsigned int fio_debug = 0;
 
 #define MAGIC1	0xa9b1c8d2
 #define MAGIC2	0xf0a1e9b3
@@ -82,3 +84,7 @@ int main(int argc, char *argv[])
 	scleanup();
 	return 0;
 }
+
+void __dprint(int type, const char *str, ...)
+{
+}
